import copy_reg
from pickle import dumps, PicklingError
import marshal, types
from types import *
import inspect
from byteplay import Code, opmap
from multiprocessing.managers import BaseProxy

from itertools import izip_longest

from .exceptions import WorkingOutsideOfContext
from .globals import _swarm_ctx_stack
from werkzeug.local import LocalProxy




#
from multiprocessing import current_process
ident = lambda:(current_process().pid, id(threading.current_thread()))
from pprint import pprint
from traceback import print_exc
import time
#

LOAD_FAST = opmap['LOAD_FAST']
STORE_FAST = opmap['STORE_FAST']
LOAD_NAME = opmap['LOAD_NAME']
STORE_NAME = opmap['STORE_NAME']
LOAD_DEREF = opmap['LOAD_DEREF']
STORE_DEREF = opmap['STORE_DEREF']

REPLACES = (STORE_DEREF, STORE_FAST, LOAD_DEREF, LOAD_FAST)

def replace_opcode(opcode):
    if opcode in (LOAD_DEREF, LOAD_FAST):
        return LOAD_NAME
    elif opcode in (STORE_DEREF, STORE_FAST):
        return STORE_NAME

def block_anonymous_code(function):
    argspec = inspect.getargspec(function)
    args = argspec.args
    if argspec.varargs is not None:
        args += argspec.varargs,
    if argspec.keywords is not None:
        args += argspec.keywords,

    code = Code.from_code(function.func_code)
    newBytecode = []
    for opcode, arg in code.code:
        if opcode in REPLACES:
            if not isinstance(arg, str) or not arg in args:
                opcode = replace_opcode(opcode)
        newBytecode.append((opcode, arg))
    code.code = newBytecode
    code.newlocals = False
    code.freevars = ()

    return code.to_code()

def code_unpickler(data):
    return marshal.loads(data)

def code_pickler(code):
    return code_unpickler, (marshal.dumps(code),)

copy_reg.pickle(CodeType, code_pickler, code_unpickler)

def module_pickler(module):
    return __import__, (module.__name__,)

copy_reg.pickle(ModuleType, module_pickler)

def method_pickler(method):
    """
    Pickle methods properly, including class methods.
    """
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    if isinstance(cls, type):
        # handle classmethods differently
        cls = obj
        obj = None
    if func_name.startswith('__') and not func_name.endswith('__'):
        #deal with mangled names
        cls_name = cls.__name__.lstrip('_')
        func_name = '_%s%s' % (cls_name, func_name)

    return method_unpickler, (func_name, obj, cls)

def method_unpickler(func_name, obj, cls):
    """
    Unpickle methods properly, including class methods.
    """
    if obj is None:
        return cls.__dict__[func_name].__get__(obj, cls)
    for cls in cls.__mro__:
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)

copy_reg.pickle(MethodType, method_pickler, method_unpickler)


class SerializedFunction(object):
    def __init__(self, name, argspec, code):
        self.name = name
        self.argspec = argspec
        self.code = code

    def __call__(self, globals={}):
        return FunctionType(self.code,
                            globals,
                            self.name,
                            self.argspec.defaults,
                            ())


BUILTINS = __builtins__.values()
NONE = object()


def serialize_namespace(namespace, names=None, omit=[]):
    for name, item in namespace.items():
        if name.startswith('__') or name in omit:
            continue

        if isinstance(item, (BuiltinMethodType,
                             BuiltinFunctionType,
                             LocalProxy)):
            continue

        if item in BUILTINS and not item in [True, False, 1, 0, None]:
            continue
        if names is None or name in names:
            try:
                yield (name, dumps(item))
            except PicklingError, e:
                if isinstance(item, FunctionType):
                    yield name, SerializedFunction(item.__name__,
                                    inspect.getargspec(item),
                                    block_anonymous_code(item))



def serialize(context, locals, block):
    if not block in context.blocks:
        context.blocks[block.__name__] = \
            dumps(block_anonymous_code(block))


    if not hasattr(context, '_stored_globals'):
        g = dict(getattr(context, '_globals', {}))
        g.update(dict(serialize_namespace(block.func_globals,
                                          omit=context.blocks)))
        context._stored_globals = g

    block_argspec = inspect.getargspec(block)
    locals_names = block_argspec.args
    if block_argspec.defaults is not None:
        defaulted = [(n,v) for (n,v) in\
                     izip_longest(reversed(locals_names),
                                  reversed(block_argspec.defaults),
                                  fillvalue=NONE) if v is not NONE]
        _locals = dict(defaulted)
        take_form_locals = [n for n in block_argspec.args if not n in _locals]
    else:
        _locals = {}
        take_form_locals = locals_names

    _locals.update(dict(serialize_namespace(locals,
                                            names=take_form_locals,
                                            )))
    return (dumps(context),
            context.blocks[block.__name__],
            block.__name__,
            block_argspec.args,
            _locals,
            context._stored_globals)

def reg_picklers():
    copy_reg.pickle(CodeType, code_pickler, code_unpickler)
    copy_reg.pickle(ModuleType, module_pickler)
    copy_reg.pickle(MethodType, method_pickler, method_unpickler)
    copy_reg.pickle(FunctionType, function_pickler, function_unpickler)