from Queue import Queue as SimpleQueue, Empty

from multiprocessing import Queue, Lock, current_process, connection
from multiprocessing.managers import SyncManager, BaseProxy, PoolProxy, RemoteError, MakeProxyType
from multiprocessing.pool import Pool, worker, TERMINATE, RUN, CLOSE
from multiprocessing.util import info, debug, Finalize

import threading
from pickle import loads, dumps, PicklingError

from types import FunctionType, CodeType
import inspect
from copy import copy
import copy_reg
import types
from collections import namedtuple
from contextlib import contextmanager
import logging
import signal

from werkzeug.utils import cached_property

from .helpers import caller_locals, cascade_reload
from .globals import swarm, _swarm_ctx_stack, _transport_ctx_stack, get_stacks
from .blocks import serialize, SerializedFunction
from .sync import Context
from .define_swarm import BLOCK_PREFIX
from .exceptions import Cancel
from swarm import signals
from .logging  import create_context_logger
from .state import send_state, restore_state

BUILTINS = __builtins__

ident = lambda:(current_process().pid, id(threading.current_thread()))
from pprint import pprint
from traceback import print_exc
import time


from threading import current_thread
class Dthread(object):
    name = 'DummyThread'
DUMMYTHREAD = Dthread()

def _current_thread(*args, **kwargs):
    try:
        return current_thread()
    except TypeError:
        return DUMMYTHREAD

threading.current_thread = _current_thread

class MergedNamespace(dict):
    def __init__(self, globals={}, locals={}):
        self._globals = globals
        self._locals = locals

    def __getitem__(self, name):
        try:
            return dict.__getitem__.__get__(self, dict)(name)
        except KeyError:
            return self._locals.get(name, self._globals.get(name))


def deserialize(namespace):
    for name, value in namespace.items():
        if isinstance(value, str):
            try:
                value = loads(value)
            except:
                continue

        if isinstance(value, SerializedFunction):
            value = value(globals=namespace)
        
        namespace[name] = value

def create_transport(context, args, kwargs):
    if not 'method' in kwargs:
        kwargs['method'] = context.object.config['DEFAULT_METHOD'].lower()
    transport = context.object.transport_class(context, *args, **kwargs)
    return transport.perform()

def runblock(task, *args, **kwargs):
    import blocks

    try:
        context, block, name, locals_order, _locals, _g = task
        context = loads(context)
        if context.object.config['SAVE_STATE']:
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        context.connect()

        context.object.reloaded_module
        context.block_name = name
        context.log.debug('Enter')
    except:
        print_exc()
        raise
    try:
        _transport_ctx_stack.push(create_transport(context, args, kwargs))
        _swarm_ctx_stack.push(context)


        _globals = copy(BUILTINS)
        _globals.update(_g)

        deserialize(_globals)
        deserialize(_locals)

        _globals.update(get_stacks())
        context._globals = MergedNamespace(_globals, _locals)

        copy_reg.pickle(NestedSwarmContext,
                        pickle_swarm_context,
                        unpickle_swarm_context)

        code = loads(block)
        args = tuple((_locals[name] for name in locals_order))
        recievers = []

        signals.context.signal('enter-block').send(context)
        if context._mode == CONTEXT.RESTORE:
            signals.context.signal('enter-restored-block').send(context)

        generator = types.FunctionType(code,
                                       _globals,
                                       name,
                                       ())(*args)


        if isinstance(generator, types.GeneratorType):
            for item in generator:
                context.items.put(item)
                context.log.debug('Send item %s'%type(item))
    except Cancel, e:
        e.log()
    except Exception, e:
        context.log.exception(name)
    finally:
        signals.context.signal('exit-block').send(context)

        _transport_ctx_stack.pop()
        with context.pool_counter_lock:
            context.pool_is_running.value -= 1
            context.log.debug('Exit')

class ContextPool(Pool):
    """
     - pool workers are non daemonic
     - task that left in inqueue redirected to caller pipe on terminate
    """

    def _repopulate_pool(self):
        """Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for i in range(self._processes - len(self._pool)):
            w = self.Process(target=worker,
                             args=(self._inqueue, self._outqueue,
                                   self._initializer,
                                   self._initargs, self._maxtasksperchild)
                            )
            self._pool.append(w)
            w.name = w.name.replace('Process', 'ContextWorker')
            # w.daemon = True
            w.start()
            debug('added worker')

    def shutdown(self, tail):
        super(ContextPool, self).close()
        terminate = False
        with self._inqueue._rlock:
            stolen = 0
            while self._inqueue._reader.poll():
                stolen += 1
                task = self._inqueue._reader.recv()
                if task is not None:
                    task_id, _ = task[0:2]
                    self._cache.pop(task_id)
                    tail.send(task[2:])
            for i in range(stolen):
                self._inqueue.put(None)
        if not self._cache:
            self.terminate()
        self.join()
        return stolen
class CONTEXT:
    RUN = 0
    RESTORE = 1

class SwarmContext(Context):
    def __init__(self, swarm, caller, *args, **kwargs):
        self.object = swarm
        self.blocks = {'define_swarm':None}
        self.caller = caller
        self.block_name = 'top'
        self.is_top = not len(_swarm_ctx_stack)

        if len(_swarm_ctx_stack):
            self._log_queue = _swarm_ctx_stack.top._log_queue

        super(SwarmContext, self).__init__(*args, **kwargs)

    def _apply_nested(self):
        while True:
            args, kwargs = self._apply.get()
            if not any((args, kwargs)):
                break
            try:
                # pprint(args)
                # pprint(kwargs)
                # print '---'
                self._pool.apply_async(*args, **kwargs)
            except AssertionError:
                self._apply.put((args, kwargs))
                break

    def _prepare_context(self):
        try:
            if self.object.reloaded_module:
                cascade_reload(self.object.reloaded_module)

            self._applier = threading.Thread(target=self._apply_nested)
            self._applier.daemon = True

            self._apply = self.Queue()

            if not hasattr(self, '_log_queue'):
                self._log_queue = self.Queue()
            self.caller.send(self._log_queue)

            self.items = self.Queue()
            self.caller.send(self.items)

            self.pool_is_running = self.Value('i', 0)
            self.caller.send(self.pool_is_running)

            self.pool_counter_lock = self.Lock()

            self.caller.send(self.pool_counter_lock)

            self._applier.start()
            if self.is_top:
                signals.context.signal('prepare-top').send(self)
                
            signals.context.signal('prepare').send(self)

        except Cancel:
            raise
        except:
            print 'On prepare context %s:%s'%(current_process(), current_thread())
            print_exc()

    def _init(self):
        self._mode = CONTEXT.RUN
        try:
            self._prepare_context()
        except Cancel, e:
            e.log()
            self.caller.recv()
            self.caller.send(None)
            return

        url, kwargs = self.caller.recv()
        try:
            swarm_func, defaults = self.object.url_adapter.match(url)

            defaults.update(kwargs)
            func = self.object.swarm_functions[swarm_func]
            maybe_generator = func(**defaults)
            if isinstance(maybe_generator, types.GeneratorType):
                for item in maybe_generator:
                    self.log.debug('Send item %r'%'%s'%item)
                    self.items.put(item)
        except Cancel, e:
            e.log()
            return #??? is it good?
        except:
            self.log.exception('On init:%s'%url)
        finally:
            self.caller.send(None)

        send_state(self)

    def _restore(self):
        self._mode = CONTEXT.RESTORE
        try:
            self._prepare_context()
        except Cancel, e:
            e.log()
            self.caller.recv()
            self.caller.send(None)
            return

        _id = self.caller.recv()

        try:
            for item in restore_state(self, _id, 'items'):
                self.log.debug('Send restored item %r'%'%s'%item)
                self.items.put(item)
            for fn, args, kwargs in restore_state(self, _id, 'tasks'):
                args = list(args)
                args[0] = list(args[0])
                args[0][0] = dumps(self)
                try:
                    signals.context.signal('apply-restored').send(self, args=args, kwargs=kwargs)
                except Cancel, e:
                    e.log()
                    continue                
                self.apply(fn, *(tuple(args), kwargs))
        except:
            self.log.exception('On restore:%s'%_id)
        finally:
            self.caller.send(None)

        send_state(self)

    def apply(self, *args, **kwargs):
        self._apply.put((args, kwargs))
        with self.pool_counter_lock:
            self.pool_is_running.value += 1

    @cached_property
    def _pool(self):
        return ContextPool(self.object.competitors)


    def __call__(self, *args, **kwargs):
        if callable(args[-1]) and args[-1].__name__.startswith(BLOCK_PREFIX):
            block, args = args[-1], args[:-1]
        else:
            block, args = args[0], args[1:]


        task = serialize(self, caller_locals(1), block)
        for task, _args, _kwargs in self.object.transport_class.iter_tasks(task, *args, **kwargs):
            try:
                signals.context.signal('apply').send(self, args=_args, kwargs=_kwargs)
            except Cancel, e:
                e.log()
                continue
            self.apply(runblock, (task,) + _args, _kwargs)

    @cached_property
    def log(self):
        if self.object.config['DEBUG']:
            self.object.log_level = logging.DEBUG
        return create_context_logger(self, self._log_queue)

class NestedSwarmContext(SwarmContext):
    def __init__(self, swarm, *args, **kwargs):
        self.object = swarm
        self.blocks = {'define_swarm': None}
        super(SwarmContext, self).__init__(*args, **kwargs)

def pickle_swarm_context(context):
    proxies = {}
    for name, proxy in context.__dict__.items():
        if name.startswith('__'):
            continue
        if not isinstance(proxy, BaseProxy):
            continue
        proxies[name] = proxy
    
    return (unpickle_swarm_context, (context.object,
                                     proxies,
                                     context._address,
                                     bytes(context._authkey),
                                     context._serializer,
                                     context._mode,))

def unpickle_swarm_context(object, proxies, address, authkey, serializer, mode):
    context = NestedSwarmContext(object, address, authkey)
    context._serializer = serializer
    context._mode = mode
    
    for (name, proxy) in proxies.items():
        setattr(context, name, proxy)
    return context


copy_reg.pickle(SwarmContext, pickle_swarm_context, unpickle_swarm_context) 