from __future__ import absolute_import

import logging

class WorkingOutsideOfContext(RuntimeError):
    pass

class PerformLoggingException(Exception):
    def log(self, *args, **kwargs):
        from . import swarm
        swarm.log.log(self.level, self.message%{'signal':self.signal.name})

class LogInfoException(PerformLoggingException):
    level = logging.INFO

class LogErrorException(PerformLoggingException):
    level = logging.ERROR

class LogDebugException(PerformLoggingException):
    level = logging.DEBUG

class SignalEmittedException(Exception):
    pass

class InfoLoggedCancel(LogInfoException, SignalEmittedException):
    pass

class ErrorLoggedCancel(LogErrorException, SignalEmittedException):
    pass

class DebugLoggedCancel(LogDebugException, SignalEmittedException):
    pass

class SilentCancel(SignalEmittedException):
    def log(self, *args, **kwargs):
        return

Cancel = (InfoLoggedCancel, SilentCancel, ErrorLoggedCancel, DebugLoggedCancel)