from functools import wraps
from weakref import WeakValueDictionary, ref
from multiprocessing import Lock

from blinker import Signal

_lock = Lock()

class LockedSignal(Signal):

    def _locked_super_method(self, name, *args, **kwargs):        
        try:
            with _lock:
                return getattr(super(LockedSignal, self), name)(*args, **kwargs)
        except Exception, e:
            e.signal = self
            raise

    def connect(self, *args, **kwargs):
        return self._locked_super_method('connect', *args, **kwargs)

    def disconnect(self, *args, **kwargs):
        return self._locked_super_method('disconnect', *args, **kwargs)

    def send(self, *args, **kwargs):
        return self._locked_super_method('send', *args, **kwargs)

class NamedSignal(LockedSignal):
    """A named generic notification emitter."""

    def __init__(self, name, doc=None):
        LockedSignal.__init__(self, doc)

        #: The name of this signal.
        self.name = name

    def __repr__(self):
        base = Signal.__repr__(self)
        return "%s; %r>" % (base[:-1], self.name)

__sigstore__ = []

class Namespace(WeakValueDictionary):
    """A mapping of signal names to signals."""

    def _wraps_signal_method(self, name, proxy=None, **kwargs):
        method = getattr(self.signal(name, **kwargs), proxy)
        @wraps(method)
        def proxy(func):
            return method(func)
        return method

    def signal(self, name, doc=None):
        """Return the :class:`NamedSignal` *name*, creating it if required.

        Repeated calls to this function will return the same signal object.

        """
        try:
            return self[name]
        except KeyError:
            signal = NamedSignal(name, doc)
            __sigstore__.append(signal)
            return self.setdefault(name, signal)

    def connect(self, name, **kwargs):
        return self._wraps_signal_method(name, proxy='connect', **kwargs)


    def disconnect(self):
        for signal in self.itervalues():
            for reciever in signal.receivers.values():
                signal.disconnect(reciever())

swarm = Namespace()
context = Namespace()
transport = Namespace()

def disconnect():
    swarm.disconnect()
    context.disconnect()
    transport.disconnect()
