from pickle import dump, load
from os import path, makedirs, listdir as ls, unlink
from shutil import rmtree
from Queue import Empty
from multiprocessing import current_process, active_children
import time


def send_state(context):
    from swarm import signals
    _id = context._process.pid
    context.caller.recv()
    context._apply.put((None, None))
    context._applier.join()
    if context.object.config['SAVE_STATE']:
        signals.context.signal('before-save-state').send(context, _id=_id)
        context.log.debug('Sending state')
        if '_pool' in context.__dict__:
            stolen = context._pool.shutdown(context.caller)
            with context.pool_counter_lock:
                context.pool_is_running.value -= stolen

        while True:
            try:
                context.caller.send(context._apply.get_nowait())
            except Empty:
                break
        signals.context.signal('after-save-state').send(context, _id=_id)
        context.log.debug('State sent')
    
    context.caller.send(None)
    context.caller.close()

def state_path(context, _id):
    return path.join(context.object.instance_dir('pending'), '%s'%_id)

def save_state(context, **kwargs):
    if not kwargs:
        return
    _id = context._process.pid
    total_entities = 0
    for name, items in kwargs.items():
        total_entities += len(items)
        state = state_path(context, _id)
        if not path.exists(state):
            makedirs(state)
        _mode = path.join(state, name)

        if not path.exists(_mode):
            makedirs(_mode)
            for i, item in enumerate(items):
                with open(path.join(_mode, str(i)), 'wb') as f:
                    dump(item, f)

    if not total_entities:
        state = state_path(context, _id)
        rmtree(state)

def restore_state(context, _id, mode):
    from swarm import signals
    signals.context.signal('before-restore-state').send(context, _id=_id, mode=mode)
    _mode = path.join(state_path(context, _id), mode)
    _items = [path.join(_mode, i) for i in sorted(ls(_mode))]

    for item in _items:
        with open(item, 'rb') as f:
            yield load(f)
        unlink(item)

    rmtree(_mode)
    if not ls(path.dirname(_mode)):
        rmtree(path.dirname(_mode))
    signals.context.signal('after-restore-state').send(context, _id=_id, mode=mode)