import threading
from Queue import Empty
from multiprocessing.managers import SyncManager, Server, State, BaseProxy
from multiprocessing import util, Queue, connection, Process, current_process


ident = lambda:(current_process().pid, id(threading.current_thread()))
from pprint import pprint
from traceback import print_exc
import time

import signal
class Server(Server):
    def serve_forever(self, writer, manager, initializer, initargs, ignore_exit=False):
        '''
        Run the server forever
        '''
        current_process()._manager_server = self
        if ignore_exit:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
        try:
            try:
                manager._state.value = State.STARTED
                manager._address = self.address
                self._initializer = threading.Thread(target=initializer,
                                                     args=initargs)
                self._initializer.daemon = True
                self._initializer.start()
                writer.close()
                while 1:
                    try:
                        c = self.listener.accept()
                    except (OSError, IOError):
                        continue
                    t = threading.Thread(target=self.handle_request, args=(c,))
                    t.daemon = True
                    t.start()


            except (KeyboardInterrupt, SystemExit):
                pass
        finally:
            self.stop = 999
            self.listener.close()

class UnterminableProcess(Process):
    # terminate of underlying SyncManager process will make top level SyncManager
    # connections broken
    def terminate(self):
        pass

class Context(SyncManager):
    _Server = Server

    @classmethod
    def _run_server(cls, manager, registry, address, authkey, serializer, writer,
                    initializer=None, initargs=(), ignore_exit=False):
        '''
        Create a server, report its address and run it
        '''
        # create server
        server = cls._Server(registry, address, authkey, serializer)

        # inform parent process of the server's address
        writer.send(server.address)
        # writer.close()
        # denz: will close bight before server loop start

        # run the manager
        util.info('manager serving at %r', server.address)
        server.serve_forever(writer, manager, initializer, initargs, ignore_exit=ignore_exit)

    def start(self, initializer=None, initargs=(), ignore_exit=False):
        '''
        Spawn a server process for this manager object
        '''
        assert self._state.value == State.INITIAL

        if initializer is not None and not hasattr(initializer, '__call__'):
            raise TypeError('initializer must be a callable')

        # pipe over which we will retrieve address of server
        reader, writer = connection.Pipe(duplex=False)

        # spawn process which runs a server
        self._process = UnterminableProcess(
            target=type(self)._run_server,
            args=(self, self._registry, self._address, self._authkey,
                  self._serializer, writer, initializer, initargs, ignore_exit),
            )
        ident = ':'.join(str(i) for i in self._process._identity)
        self._process.name = type(self).__name__  + '-' + ident
        self._process.start()
        self._address = reader.recv()

        # register a finalizer
        self._state.value = State.STARTED
        self.shutdown = util.Finalize(
            self, type(self)._finalize_manager,
            args=(self._process, self._address, self._authkey,
                  self._state, self._Client),
            exitpriority=0
            )
        writer.close()
        reader.close()